/*
    semu - Sega Saturn Emulator
    Copyright (C) 1998  Chuck Mason

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Chuck Mason IV                      chuckjr@sundial.net
    477 Moffat Loop
    Oviedo, Fl 32765
*/

#include <config.h>
#include <stdio.h>
#include <stdlib.h>
#include "dyninstr.h"
#include "opcodes.h"
#include "formats.h"

static instruction_list sh2_instruction_list[] = {
 { 0xFFFF, 0x0009, X_FORMAT, sh2_opcode_nop, "NOP" }, /* NOP */
 { 0xF00F, 0x300C, NM_FORMAT, sh2_opcode_add, "ADD R%d, R%d" }, /* ADD Rm, Rn */
 { 0xF000, 0x7000, NI_FORMAT, sh2_opcode_addi, "ADD %02X, R%d" }, /* ADD #imm, Rn */
 { 0xF00F, 0x300E, NM_FORMAT, sh2_opcode_addc, "ADDC R%d, R%d" }, /* ADDC Rm, Rn */
 { 0xF00F, 0x300F, NM_FORMAT, sh2_opcode_addv, "ADDV R%d, R%d" }, /* ADDV Rm, Rn */
 { 0xF00F, 0x2009, NM_FORMAT, sh2_opcode_and, "AND R%d, R%d" }, /* AND Rm, Rn */
 { 0xFF00, 0xC900, I_FORMAT, sh2_opcode_andi, "AND %02X, R0" }, /* AND #imm, R0 */
 { 0xFF00, 0xCD00, I_FORMAT, sh2_opcode_andm, "AND.B %02X, @(R0, GBR)" }, /* AND.B #imm, @(R0, GBR) */
 { 0xFF00, 0x8B00, D_FORMAT, sh2_opcode_bf, "BF %08X" }, /* BF TARGT_F */
 { 0xFF00, 0x8F00, D_FORMAT, sh2_opcode_bfs, "BF/S %08X" }, /* BF/S TARGT_F */
 { 0xFF00, 0x8900, D_FORMAT, sh2_opcode_bt, "BT %08X" }, /* BT TARGT_T */
 { 0xFF00, 0x8D00, D_FORMAT, sh2_opcode_bts, "BT/S %08X" }, /* BT/S TARGT_T */
 { 0xFFFF, 0x0028, X_FORMAT, sh2_opcode_clrmac, "CLRMAC" }, /* CLRMAC */

 { 0xFFFF, 0x0008, X_FORMAT, sh2_opcode_clrt, "CLRT" }, /* CLRT */
 { 0xFFFF, 0x0018, X_FORMAT, sh2_opcode_sett, "SETT" }, /* SETT */
 
 /* CMP/xx Series */
 { 0xF00F, 0x3000, NM_FORMAT, sh2_opcode_cmpeq, "CMP/EQ R%d, R%d" }, /* CMP/EQ Rm, Rn */
 { 0xF00F, 0x3003, NM_FORMAT, sh2_opcode_cmpge, "CMP/GE R%d, R%d" }, /* CMP/GE Rm, Rn */
 { 0xF00F, 0x3007, NM_FORMAT, sh2_opcode_cmpgt, "CMP/GT R%d, R%d" }, /* CMP/GT Rm, Rn */
 { 0xF00F, 0x3006, NM_FORMAT, sh2_opcode_cmphi, "CMP/HI R%d, R%d" }, /* CMP/HI Rm, Rn */
 { 0xF00F, 0x3002, NM_FORMAT, sh2_opcode_cmphs, "CMP/HS R%d, R%d" }, /* CMP/HS Rm, Rn */
 { 0xF0FF, 0x4015, N_FORMAT, sh2_opcode_cmppl,  "CMP/PL R%d" }, /* CMP/PL Rn */
 { 0xF0FF, 0x4011, N_FORMAT, sh2_opcode_cmppz,  "CMP/PZ R%d" }, /* CMP/PZ Rn */
 { 0xF00F, 0x200C, NM_FORMAT, sh2_opcode_cmpstr, "CMP/STR R%d, R%d" }, /* CMP/STR Rm, Rn */
 { 0xFF00, 0x8800, I_FORMAT, sh2_opcode_cmpim, "CMP/IM %02X, R0" }, /* CMP/IM #imm, R0 */

 /* DIV{0S,0U,1} series */
 { 0xF00F, 0x2007, NM_FORMAT, sh2_opcode_div0s, "DIV0S R%d, R%d" }, /* DIV0S Rm, Rn */
 { 0xFFFF, 0x0019, X_FORMAT, sh2_opcode_div0u, "DIV0U" }, /* DIV0U */
 { 0xF00F, 0x3004, NM_FORMAT, sh2_opcode_div1, "DIV1 R%d, R%d" }, /* DIV1 Rm, Rn */

 { 0xF00F, 0x300D, NM_FORMAT, sh2_opcode_dmuls, "DMULS.L R%d, R%d" }, /* DMULS.L Rm, Rn */ 
 { 0xF00F, 0x3005, NM_FORMAT, sh2_opcode_dmulu, "DMULU.L R%d, R%d" }, /* DMULU.L Rm, Rn */

 { 0xF0FF, 0x4010, N_FORMAT, sh2_opcode_dt, "DT R%d" }, /* DT Rn */

 { 0xF00F, 0x600E, NM_FORMAT, sh2_opcode_extsb, "EXTS.B R%d, R%d" }, /* EXTS.B Rm, Rn */
 { 0xF00F, 0x600F, NM_FORMAT, sh2_opcode_extsw, "EXTS.W R%d, R%d" }, /* EXTS.W Rm, Rn */
 { 0xF00F, 0x600C, NM_FORMAT, sh2_opcode_extub, "EXTU.B R%d, R%d" }, /* EXTU.B Rm, Rn */
 { 0xF00F, 0x600D, NM_FORMAT, sh2_opcode_extuw, "EXTU.W R%d, R%d" }, /* EXTU.W Rm, Rn */ 

 { 0xFFFF, 0x001B, X_FORMAT, sh2_opcode_sleep, "SLEEP" }, /* SLEEP */ 
 { 0xF000, 0xA000, D12_FORMAT, sh2_opcode_bra, "BRA %03X" }, /* BRA label */
 { 0xF0FF, 0x0023, M_FORMAT, sh2_opcode_braf, "BRA R%d"  }, /* BRA Rm */ 
 { 0xF000, 0xB000, D12_FORMAT, sh2_opcode_bsr, "BSR %03X" }, /* BSR label */
 { 0xF0FF, 0x0003, M_FORMAT, sh2_opcode_bsrf, "BSRF R%d"  }, /* BSRF Rm */

 { 0xF0FF, 0x402B, M_FORMAT, sh2_opcode_jmp, "JMP @R%d" }, /* JMP @Rm */
 { 0xF0FF, 0x400B, M_FORMAT, sh2_opcode_jsr, "JSR @R%d" }, /* JSR @Rm */

 /* LDC series */
 { 0xF0FF, 0x400E, M_FORMAT, sh2_opcode_ldcsr, "LDC R%d, SR" }, /* LDC Rm, SR */
 { 0xF0FF, 0x401E, M_FORMAT, sh2_opcode_ldcgbr, "LDC R%d, GBR" }, /* LDC Rm, GBR */
 { 0xF0FF, 0x402E, M_FORMAT, sh2_opcode_ldcvbr, "LDC R%d, VBR" }, /* LDC Rm, VBR */
 { 0xF0FF, 0x4007, M_FORMAT, sh2_opcode_ldcmsr, "LDC.L @R%d+, SR" }, /* LDC.L @Rm+, SR */
 { 0xF0FF, 0x4017, M_FORMAT, sh2_opcode_ldcmgbr, "LDC.L @R%d+, GBR" }, /* LDC.L @Rm+, GBR */
 { 0xF0FF, 0x4027, M_FORMAT, sh2_opcode_ldcmvbr, "LDC.L @R%d+, VBR" }, /* LDC.L @Rm+, VBR */
 /* ---------- */
 /* LDS series */ 
 { 0xF0FF, 0x400A, M_FORMAT, sh2_opcode_ldsmach, "LDS R%d, MACH" }, /* LDS Rm, MACH */
 { 0xF0FF, 0x401A, M_FORMAT, sh2_opcode_ldsmacl, "LDS R%d, MACL" }, /* LDS Rm, MACL */
 { 0xF0FF, 0x402A, M_FORMAT, sh2_opcode_ldspr, "LDS R%d, PR"   }, /* LDS Rm, PR */
 { 0xF0FF, 0x4006, M_FORMAT, sh2_opcode_ldsmmach, "LDS.L @R%d+, MACH" }, /* LDS.L @Rm+, MACH */
 { 0xF0FF, 0x4016, M_FORMAT, sh2_opcode_ldsmmacl, "LDS.L @R%d+, MACL" }, /* LDS.L @Rm+, MACL */
 { 0xF0FF, 0x4026, M_FORMAT, sh2_opcode_ldsmpr,   "LDS.L @R%d+, PR" }, /* LDS.L @Rm+, PR */
 /* ---------- */
 { 0xF00F, 0x000F, NM_FORMAT, sh2_opcode_macl, "MAC.L @R%d+, @R%d+" }, /* MAC.L @Rm+, @Rn+ */
 { 0xF00F, 0x400F, NM_FORMAT, sh2_opcode_macw, "MAC.W @R%d+, @R%d+" }, /* MAC.W @Rm+, @Rn+ */
 /* ---------- */
 /* MOV series */
 { 0xF00F, 0x6003, NM_FORMAT, sh2_opcode_mov, "MOV R%d, R%d" }, /* MOV Rm, Rn */

 { 0xF00F, 0x0004, NM_FORMAT, sh2_opcode_movbs0, "MOV.B R%d, @(R0, R%d)" }, /* MOV.B Rm, @(R0, Rn) */
 { 0xF00F, 0x0005, NM_FORMAT, sh2_opcode_movws0, "MOV.W R%d, @(R0, R%d)" }, /* MOV.W Rm, @(R0, Rn) */
 { 0xF00F, 0x0006, NM_FORMAT, sh2_opcode_movls0, "MOV.L R%d, @(R0, R%d)" }, /* MOV.L Rm, @(R0, Rn) */

 { 0xF00F, 0x2000, NM_FORMAT, sh2_opcode_movbs, "MOV.B R%d, @R%d" }, /* MOV.B Rm, @Rn */
 { 0xF00F, 0x2001, NM_FORMAT, sh2_opcode_movws, "MOV.W R%d, @R%d" }, /* MOV.W Rm, @Rn */
 { 0xF00F, 0x2002, NM_FORMAT, sh2_opcode_movls, "MOV.L R%d, @R%d" }, /* MOV.L Rm, @Rn */

 { 0xF00F, 0x6000, NM_FORMAT, sh2_opcode_movbl, "MOV.B @R%d, R%d" }, /* MOV.B @Rm, Rn */
 { 0xF00F, 0x6001, NM_FORMAT, sh2_opcode_movwl, "MOV.W @R%d, R%d" }, /* MOV.W @Rm, Rn */
 { 0xF00F, 0x6002, NM_FORMAT, sh2_opcode_movll, "MOV.L @R%d, R%d" }, /* MOV.L @Rm, Rn */

 { 0xF00F, 0x2004, NM_FORMAT, sh2_opcode_movbm, "MOV.B R%d, @-R%d" }, /* MOV.B Rm, @-Rn */
 { 0xF00F, 0x2005, NM_FORMAT, sh2_opcode_movwm, "MOV.W R%d, @-R%d" }, /* MOV.W Rm, @-Rn */
 { 0xF00F, 0x2006, NM_FORMAT, sh2_opcode_movlm, "MOV.L R%d, @-R%d" }, /* MOV.L Rm, @-Rn */

 { 0xF00F, 0x6004, NM_FORMAT, sh2_opcode_movbp, "MOV.B @R%d+, R%d" }, /* MOV.B @Rm+, Rn */
 { 0xF00F, 0x6005, NM_FORMAT, sh2_opcode_movwp, "MOV.W @R%d+, R%d" }, /* MOV.W @Rm+, Rn */
 { 0xF00F, 0x6006, NM_FORMAT, sh2_opcode_movlp, "MOV.L @R%d+, R%d" }, /* MOV.L @Rm+, Rn */
 
 { 0xF00F, 0x000C, NM_FORMAT, sh2_opcode_movbl0, "MOV.B @(R0, R%d), R%d" }, /* MOV.B @(R0, Rm), Rn */
 { 0xF00F, 0x000D, NM_FORMAT, sh2_opcode_movwl0, "MOV.W @(R0, R%d), R%d" }, /* MOV.W @(R0, Rm), Rn */
 { 0xF00F, 0x000E, NM_FORMAT, sh2_opcode_movll0, "MOV.L @(R0, R%d), R%d" }, /* MOV.L @(R0, Rm), Rn */
 
 { 0xF000, 0xE000, NI_FORMAT, sh2_opcode_movi, "MOV %02X, R%d" },  /* MOV #imm, Rn */
 { 0xF000, 0x9000, ND8_FORMAT, sh2_opcode_movwi, "MOV.W @(%08X, PC), R%d" }, /* MOV.W @(disp, PC), Rn */
 { 0xF000, 0xD000, ND8_FORMAT, sh2_opcode_movli, "MOV.L @(%08X, PC), R%d" }, /* MOV.L @(disp, PC), Rn */

 { 0xFF00, 0xC400, D_FORMAT, sh2_opcode_movblg, "MOV.B @(%02X, GBR), R0" }, /* MOV.B @(disp, GBR), R0 */
 { 0xFF00, 0xC500, D_FORMAT, sh2_opcode_movwlg, "MOV.W @(%03X, GBR), R0" }, /* MOV.W @(disp, GBR), R0 */
 { 0xFF00, 0xC600, D_FORMAT, sh2_opcode_movllg, "MOV.L @(%03X, GBR), R0" }, /* MOV.L @(disp, GBR), R0 */
 { 0xFF00, 0xC000, D_FORMAT, sh2_opcode_movbsg, "MOV.B R0, @(%02X, GBR)" }, /* MOV.B R0, @(disp, GBR) */
 { 0xFF00, 0xC100, D_FORMAT, sh2_opcode_movwsg, "MOV.W R0, @(%03X, GBR)" }, /* MOV.W R0, @(disp, GBR) */
 { 0xFF00, 0xC200, D_FORMAT, sh2_opcode_movlsg, "MOV.L R0, @(%03X, GBR)" }, /* MOV.L R0, @(disp, GBR) */
 
 { 0xFF00, 0x8000, ND4_FORMAT, sh2_opcode_movbs4, "MOV.B R0, @(%01X, R%d)" },
 { 0xFF00, 0x8100, ND4_FORMAT, sh2_opcode_movws4, "MOV.W R0, @(%01X, R%d)" },
 { 0xF000, 0x1000, NMD_FORMAT, sh2_opcode_movls4, "MOV.L R%d, @(%01X, R%d)" },
 { 0xFF00, 0x8400, MD_FORMAT, sh2_opcode_movbl4, "MOV.B @(%01X, R%d), R0" },
 { 0xFF00, 0x8500, MD_FORMAT, sh2_opcode_movwl4, "MOV.W @(%01X, R%d), R0" },
 { 0xF000, 0x5000, NMD1_FORMAT, sh2_opcode_movll4, "MOV.L @(%01X, R%d), R%d" },

 { 0xFF00, 0xC700, D_FORMAT, sh2_opcode_mova, "MOVA @(%08X, PC), R0" },  
 { 0xF0FF, 0x0029, N_FORMAT, sh2_opcode_movt, "MOVT R%d" },
 /* NO MORE MOVs!! */
 { 0xF00F, 0x0007, NM_FORMAT, sh2_opcode_mull, "MUL.L R%d, R%d" },
 { 0xF00F, 0x200F, NM_FORMAT, sh2_opcode_muls, "MULS(.W) R%d, R%d" },
 { 0xF00F, 0x200E, NM_FORMAT, sh2_opcode_mulu, "MULU(.W) R%d, R%d" },

 { 0xF00F, 0x600B, NM_FORMAT, sh2_opcode_neg, "NEG R%d, R%d" },
 { 0xF00F, 0x600A, NM_FORMAT, sh2_opcode_negc, "NEGC R%d, R%d" },
 
 { 0xF00F, 0x6007, NM_FORMAT, sh2_opcode_not, "NOT R%d, R%d" },
 
 { 0xF00F, 0x200B, NM_FORMAT, sh2_opcode_or, "OR R%d, R%d" },
 { 0xFF00, 0xCB00, I_FORMAT, sh2_opcode_ori, "OR %02X, R0" },
 { 0xFF00, 0xCF00, I_FORMAT, sh2_opcode_orm, "OR.B %02X, @(R0, GBR)" },
 
 { 0xF0FF, 0x4024, N_FORMAT, sh2_opcode_rotcl, "ROTCL R%d" },
 { 0xF0FF, 0x4025, N_FORMAT, sh2_opcode_rotcr, "ROTCR R%d" }, 
 { 0xF0FF, 0x4004, N_FORMAT, sh2_opcode_rotl,  "ROTL R%d" },
 { 0xF0FF, 0x4005, N_FORMAT, sh2_opcode_rotr,  "ROTR R%d" },
 
 { 0xFFFF, 0x002B, X_FORMAT, sh2_opcode_rte, "RTE" },
 { 0xFFFF, 0x000B, X_FORMAT, sh2_opcode_rts, "RTS" },

 { 0xF0FF, 0x4020, N_FORMAT, sh2_opcode_shal, "SHAL R%d" },
 { 0xF0FF, 0x4021, N_FORMAT, sh2_opcode_shar, "SHAR R%d" },
 { 0xF0FF, 0x4000, N_FORMAT, sh2_opcode_shal, "SHLL R%d" },
 
 { 0xF0FF, 0x4008, N_FORMAT, sh2_opcode_shll2, "SHLL2 R%d" },
 { 0xF0FF, 0x4018, N_FORMAT, sh2_opcode_shll8, "SHLL8 R%d" },
 { 0xF0FF, 0x4028, N_FORMAT, sh2_opcode_shll16, "SHLL16 R%d" },
 
 { 0xF0FF, 0x4001, N_FORMAT, sh2_opcode_shlr, "SHLR R%d" },
 
 { 0xF0FF, 0x4009, N_FORMAT, sh2_opcode_shlr2, "SHLR2 R%d" },
 { 0xF0FF, 0x4019, N_FORMAT, sh2_opcode_shlr8, "SHLR8 R%d" },
 { 0xF0FF, 0x4029, N_FORMAT, sh2_opcode_shlr16, "SHLR16 R%d" },
 
 { 0xF0FF, 0x0002, N_FORMAT, sh2_opcode_stcsr, "STC SR, R%d" },
 { 0xF0FF, 0x0012, N_FORMAT, sh2_opcode_stcgbr, "STC GBR, R%d" },
 { 0xF0FF, 0x0022, N_FORMAT, sh2_opcode_stcvbr, "STC VBR, R%d" },
 { 0xF0FF, 0x4003, N_FORMAT, sh2_opcode_stcmsr, "STC.L SR, @-R%d" },
 { 0xF0FF, 0x4013, N_FORMAT, sh2_opcode_stcmgbr, "STC.L GBR, @-R%d" },
 { 0xF0FF, 0x4023, N_FORMAT, sh2_opcode_stcmvbr, "STC.L VBR, @-R%d" },
 
 { 0xF0FF, 0x000A, N_FORMAT, sh2_opcode_stsmach, "STS MACH, R%d" },
 { 0xF0FF, 0x001A, N_FORMAT, sh2_opcode_stsmacl, "STS MACL, R%d" },
 { 0xF0FF, 0x002A, N_FORMAT, sh2_opcode_stspr, "STS PR, R%d" },
 { 0xF0FF, 0x4002, N_FORMAT, sh2_opcode_stsmmach, "STS.L MACH, @-R%d" },
 { 0xF0FF, 0x4012, N_FORMAT, sh2_opcode_stsmmacl, "STS.L MACL, @-R%d" },
 { 0xF0FF, 0x4022, N_FORMAT, sh2_opcode_stsmpr, "STS.L PR, @-R%d" },

 { 0xF00F, 0x3008, NM_FORMAT, sh2_opcode_sub, "SUB R%d, R%d" },
 { 0xF00F, 0x300A, NM_FORMAT, sh2_opcode_subc, "SUBC R%d, R%d" },
 { 0xF00F, 0x300B, NM_FORMAT, sh2_opcode_subv, "SUBV R%d, R%d" }, 
 
 { 0xF00F, 0x6008, NM_FORMAT, sh2_opcode_swapb, "SWAP.B R%d, R%d" },
 { 0xF00F, 0x6009, NM_FORMAT, sh2_opcode_swapw, "SWAP.W R%d, R%d" },
 
 { 0xF0FF, 0x401B, N_FORMAT, sh2_opcode_tas, "TAS.B @R%d" },
 
 { 0xFF00, 0xC300, I_FORMAT, sh2_opcode_trapa, "TRAPA %02X" },
 
 { 0xF00F, 0x2008, NM_FORMAT, sh2_opcode_tst, "TST R%d, R%d" },
 { 0xFF00, 0xC800, I_FORMAT, sh2_opcode_tsti, "TST %02X, R0" },
 { 0xFF00, 0xCC00, I_FORMAT, sh2_opcode_tstm, "TST.B %02X, (R0, GBR)" },
 
 { 0xF00F, 0x200A, NM_FORMAT, sh2_opcode_xor, "XOR R%d, R%d" },
 { 0xFF00, 0xCA00, I_FORMAT, sh2_opcode_xori, "XOR %02X, R0" },
 { 0xFF00, 0xCE00, I_FORMAT, sh2_opcode_xorm, "XOR.B %02X, (R0, GBR)" },

 { 0xF00F, 0x200D, NM_FORMAT, sh2_opcode_xtrct, "XTRCT R%d, R%d" } /* yay */ 
};

static const int TOTAL_OPCODE_COUNT = 142;
static const int OPCODE_COUNT = sizeof(sh2_instruction_list) / sizeof(instruction_list);

instruction_list sh2_instructions[65536];

instruction_list sh2_find_opcode(unsigned long opcode)
{
	uint32_t x;
	instruction_list bad = { 0, 0, 0, sh2_opcode_invalid, "UNIMPLEMENTED/INVALID" };
	
	for(x = 0; x != OPCODE_COUNT; x++) {
		if((sh2_instruction_list[x].mask & opcode) == sh2_instruction_list[x].opcode)
			return sh2_instruction_list[x];
	}
	return bad;
}
		
void sh2_build_instruction_list()
{
	uint32_t optest;
	unsigned good, bad;

	for(optest = 0; optest <= 0xFFFF; optest++) 
		sh2_instructions[optest] = sh2_find_opcode(optest);
	
	for(optest = 0, good = 0, bad = 0; optest <= 0xFFFF; optest++) {
		if(sh2_instructions[optest].execute != sh2_opcode_invalid)
			good++;
		else
			bad++;
	}

	printf("SH2: %d/%d implemented opcodes.\n", OPCODE_COUNT, TOTAL_OPCODE_COUNT);
	printf("SH2: %d unimplemented opcodes.\n", TOTAL_OPCODE_COUNT - OPCODE_COUNT);
	printf("SH2: %d/65536 instructions available.\n", good);
	printf("SH2: %d unimplemented accesses.\n", bad);

}
