/*
    semu - Sega Saturn Emulator
    Copyright (C) 1998  Chuck Mason

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Chuck Mason IV                      chuckjr@sundial.net
    477 Moffat Loop
    Oviedo, Fl 32765
*/
                                                                
#include <stdio.h>
#include <stdlib.h>
#include "memory.h"
#include <vdp1.h>

uint32_t access_read_100000(sh2_processor *sh2, uint32_t addr, mem_access type)
{
	printf("Read access 1xxxxx %5X.%c (PC=%8X)\n", addr, "BWL"[type], sh2->PC);
	return 0;
}

void access_write_100000(sh2_processor *sh2, uint32_t addr, uint32_t val, mem_access type)
{
	printf("Write access 1xxxxx %5X.%c (%X) (PC=%8X)\n", addr, "BWL"[type], val, sh2->PC);
}

uint32_t access_read_200000(sh2_processor *sh2, uint32_t addr,  mem_access type)
{
	printf("Read access 2xxxxx %5X.%c (PC=%8X)\n", addr, "BWL"[type], sh2->PC);
	return 0;
}

void access_write_200000(sh2_processor *sh2, uint32_t addr, uint32_t val, mem_access type)
{
	if(val!=0)
		printf("Write access 2xxxxx %5X.%c (%X) (PC=%8X)\n", addr, "BWL"[type], val, sh2->PC);
}


static uint8_t ram[1024*1024*3/2];
static uint8_t rom[512*1024];

uint32_t sh2_read_memory(sh2_processor *sh2, uint32_t addr, mem_access type)
{
	uint32_t addr1 = addr & 0xffffff;
	switch(addr >> 27) {
	case 0:
	case 4:
		switch((addr >> 24) & 7) {
		case 0:
			if(addr1<0x80000)
				return direct_read(rom+addr1, type);
			
			if(addr1>=0x100000 && addr1<0x180000)
				return access_read_100000(sh2, addr1 & 0x7ffff, type);
			break;
		case 2:
		case 3:
			return access_read_cs0a(sh2, addr & 0x1ffffff, type);
		case 4:
			return access_read_cs1a(sh2, addr1 & 0xffffff, type);
		case 5:
			switch((addr >> 20) & 15) {
			case 0:
			case 1:
			case 2:
			case 3:
			case 4:
			case 5:
			case 6:
			case 7:
				return access_read_cs2a(sh2, addr1, type);
			case 8:
				return access_read_cs3a(sh2, addr & 0xfffff, type);
			case 9:
				return direct_read(ram+(addr & 0xfffff), type);
			case 0xa:
			case 0xb:
				return access_read_sound(sh2, addr & 0x1fffff, type);
			case 0xc: {
				int zone = (addr >> 16) & 15;
				if(zone<8)
					return access_read_vdp1_vram(sh2, addr & 0x7ffff, type);
				else if(zone<0xc)
					return access_read_vdp1_fb(sh2, addr & 0x7ffff, type);
			}
			case 0xd:
				return access_read_vdp1_regs(sh2, addr & 0xfffff, type);
			case 0xe:
				return access_read_vdp2_vram(sh2, addr & 0xfffff, type);
			case 0xf: {
				int zone = (addr >> 16) & 15;
				if(zone<8)
					return access_read_vdp2_color(sh2, addr & 0x7ffff, type);
				else if(zone<0xc)
					return access_read_vdp2_regs(sh2, addr & 0x1ffff, type);
				else if(zone<0xe)
					return access_read_smpc(sh2, addr & 0x1ffff, type);
				else
					return access_read_scu(sh2, addr & 0x1ffff, type);
			}
			}
			break;
		case 6:
		case 7:
			return direct_read(ram+addr1, type);
		}
		break;
	case 8:
	case 12:
	case 24:
		return 0;
	case 31:
		return 0;
	}
	printf("Undecoded read access at %8X.%c (PC=%8X)\n", addr, "BWL"[type], sh2->PC);
	return 0;
}

void sh2_write_memory(sh2_processor *sh2, uint32_t addr, uint32_t val, mem_access type)
{
	uint32_t addr1 = addr & 0xffffff;
	switch(addr >> 27) {
	case 0:
	case 4:
		switch((addr >> 24) & 7) {
		case 0:
			if(addr1<0x80000) {
				printf("Write access to rom %5X.%c (%X) (PC=%8X)\n", addr1, "BWL"[type], val, sh2->PC);
				return;
			}
			if(addr1>=0x100000 && addr1<0x180000) {
				access_write_100000(sh2, addr1 & 0x7ffff, val, type);
				return;
			}
			if(addr1>=0x200000 && addr1<0x300000) {
				access_write_200000(sh2, addr1 & 0xfffff, val, type);
				return;
			}
			break;
		case 2:
		case 3:
			access_write_cs0a(sh2, addr & 0x1ffffff, val, type);
			return;
		case 4:
			access_write_cs1a(sh2, addr1, val, type);
			return;
		case 5:
			switch((addr >> 20) & 15) {
			case 0:
			case 1:
			case 2:
			case 3:
			case 4:
			case 5:
			case 6:
			case 7:
				access_write_cs2a(sh2, addr1, val, type);
				return;
			case 8:
				access_write_cs3a(sh2, addr & 0xfffff, val, type);
				return;
			case 9:
				direct_write(ram+(addr & 0xfffff), val, type);
				return;
			case 0xa:
			case 0xb:
				access_write_sound(sh2, addr & 0x1fffff, val, type);
				return;
			case 0xc: {
				int zone = (addr >> 16) & 15;
				if(zone<8)
					access_write_vdp1_vram(sh2, addr & 0x7ffff, val, type);
				else if(zone<0xc)
					access_write_vdp1_fb(sh2, addr & 0x7ffff, val, type);
				else
					break;
				return;
			}
			case 0xd:
				access_write_vdp1_regs(sh2, addr & 0xfffff, val, type);
				return;
			case 0xe:
				access_write_vdp2_vram(sh2, addr & 0xfffff, val, type);
				return;
			case 0xf: {
				int zone = (addr >> 16) & 15;
				if(zone<8)
					access_write_vdp2_color(sh2, addr & 0x7ffff, val, type);
				else if(zone<0xc)
					access_write_vdp2_regs(sh2, addr & 0x1ffff, val, type);
				else if(zone<0xe)
					access_write_smpc(sh2, addr & 0x1ffff, val, type);
				else
					access_write_scu(sh2, addr & 0x1ffff, val, type);
				return;
			}
			}
			break;
		case 6:
		case 7:
			direct_write(ram+addr1, val, type);
			return;
		}
		break;
	case 8:
	case 12:
	case 24:
		return;
	case 31:
		return;
	}
	printf("Undecoded write access at %8X.%c (%X) (PC=%8X)\n", addr, "BWL"[type], val, sh2->PC);
}

void sh2_write_byte_rom(uint32_t addr, uint8_t val)
{
	rom[addr] = val;
}

