/*
    semu - Sega Saturn Emulator
    Copyright (C) 1998  Chuck Mason

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Chuck Mason IV                      chuckjr@sundial.net
    477 Moffat Loop
    Oviedo, Fl 32765
*/

#include <config.h>
#include <stdio.h>
#include "formats.h"
#include "sh2proc.h"
#include "opcodes.h"
#include "dyninstr.h"
#include "memory.h"

/* SR bits accessors */

static inline i_get_M(sh2_processor *sh2)
{
	return (sh2->SR & 0x0200) != 0;
}

static inline i_get_Q(sh2_processor *sh2)
{
	return (sh2->SR & 0x0100) != 0;
}

static inline i_get_S(sh2_processor *sh2)
{
	return (sh2->SR & 0x0002) != 0;
}

static inline i_get_T(sh2_processor *sh2)
{
	return (sh2->SR & 0x0001) != 0;
}

static inline void i_set_M(sh2_processor *sh2, int v)
{
	if(v)
		sh2->SR |= 0x0200;
	else
		sh2->SR &= ~0x0200;
}

static inline void i_set_Q(sh2_processor *sh2, int v)
{
	if(v)
		sh2->SR |= 0x0100;
	else
		sh2->SR &= ~0x0100;
}

static inline void i_set_S(sh2_processor *sh2, int v)
{
	if(v)
		sh2->SR |= 0x0002;
	else
		sh2->SR &= ~0x0002;
}

static inline void i_set_T(sh2_processor *sh2, int v)
{
	if(v)
		sh2->SR |= 0x0001;
	else
		sh2->SR &= ~0x0001;
}


int get_M(sh2_processor *sh2)
{
	return i_get_M(sh2);
}

int get_Q(sh2_processor *sh2)
{
	return i_get_Q(sh2);
}

int get_S(sh2_processor *sh2)
{
	return i_get_S(sh2);
}

int get_T(sh2_processor *sh2)
{
	return i_get_T(sh2);
}

void set_M(sh2_processor *sh2, int v)
{
	i_set_M(sh2, v);
}

void set_Q(sh2_processor *sh2, int v)
{
	i_set_Q(sh2, v);
}

void set_S(sh2_processor *sh2, int v)
{
	i_set_S(sh2, v);
}

void set_T(sh2_processor *sh2, int v)
{
	i_set_T(sh2, v);
}

/* Instruction fields accessors */
static inline uint16_t i_inst_0_3(uint16_t opcode)
{
	return (opcode & 0x000f);
}

static inline uint16_t i_inst_0_7(uint16_t opcode)
{
	return (opcode & 0x00ff);
}

static inline uint16_t i_inst_0_11(uint16_t opcode)
{
	return (opcode & 0x0fff);
}

static inline uint16_t i_inst_4_7(uint16_t opcode)
{
	return (opcode >> 4) & 0xf;
}

static inline uint16_t i_inst_8_11(uint16_t opcode)
{
	return (opcode >> 8) & 0xf;
}


static inline void delay_slot(sh2_processor *sh2, uint32_t npc)
{
	sh2->delay_instruction = sh2->PC;
	sh2->PC = npc;
	sh2->delay_slot_in_progress = 2;
}

void sh2_execute_instruction(sh2_processor *sh2)
{
	sh2->inst = sh2_read_word(sh2, sh2->PC);

	sh2_instructions[sh2->inst].execute(sh2);

	if((sh2->delay_slot_in_progress) && (--sh2->delay_slot_in_progress) == 0) {
		sh2->PC = sh2->delay_instruction;
		sh2->delay_slot_in_progress = 0;
	}
}

#ifdef ENABLE_DEBUG
static void sh2_format_instruction_string(uint16_t opcode, char *string, uint32_t adr)
{
	int32_t l;
	
	switch(sh2_instructions[opcode].format) {
		case X_FORMAT:
			strcpy(string, sh2_instructions[opcode].instrformat);
			break;
		case N_FORMAT:
		case M_FORMAT:
			sprintf(string, 
				sh2_instructions[opcode].instrformat, 
				i_inst_8_11(opcode));
			break;
		case NM_FORMAT:
			sprintf(string, 
				sh2_instructions[opcode].instrformat,
				i_inst_4_7(opcode),
				i_inst_8_11(opcode));
			break;
		case MD_FORMAT:
		case ND4_FORMAT:
			sprintf(string,
				sh2_instructions[opcode].instrformat,
				i_inst_0_3(opcode),
				i_inst_4_7(opcode));
			break;
		case NMD_FORMAT:
			sprintf(string,
				sh2_instructions[opcode].instrformat,
				i_inst_4_7(opcode),
				i_inst_0_3(opcode),
				i_inst_8_11(opcode));
			break;
		case NMD1_FORMAT:
			sprintf(string,
				sh2_instructions[opcode].instrformat,
				i_inst_0_3(opcode),
				i_inst_4_7(opcode),
				i_inst_8_11(opcode));
			break;
		case D_FORMAT:
			l = i_inst_0_7(opcode);
			if(sh2_instructions[opcode].execute == sh2_opcode_movblg ||
			   sh2_instructions[opcode].execute == sh2_opcode_movbsg) {
				/* Nothing to do */
			} else if(sh2_instructions[opcode].execute == sh2_opcode_movwlg ||
				  sh2_instructions[opcode].execute == sh2_opcode_movwsg) {
				l *= 2;
			} else if(sh2_instructions[opcode].execute == sh2_opcode_movllg ||
				sh2_instructions[opcode].execute == sh2_opcode_movlsg) {
				l *= 4;
			} else if(sh2_instructions[opcode].execute == sh2_opcode_mova) {
				l = (l*4+4+adr) & ~3;
			} else {
				if(l >= 0x80)
					l -= 0x100;
				l = l*2+4+adr;
			}

			sprintf(string,
				sh2_instructions[opcode].instrformat,
				l);
			break;
		case D12_FORMAT:
			l = i_inst_0_11(opcode);

			if(l >= 0x800)
				l -= 0x1000;
			
			sprintf(string,
				sh2_instructions[opcode].instrformat,
				l*2+4+adr);
			break;
		case ND8_FORMAT:
			l = i_inst_0_7(opcode);

			if(sh2_instructions[opcode].execute == sh2_opcode_movli)
				l = (l*4+4+adr) & 0xFFFFFFFC;
			else
				l = l*2+4+adr;

			sprintf(string,
				sh2_instructions[opcode].instrformat,
				l,
				i_inst_8_11(opcode));
			break;
		case I_FORMAT:
			sprintf(string,
				sh2_instructions[opcode].instrformat,
				i_inst_0_7(opcode));
			break;
		case NI_FORMAT:
			sprintf(string,
				sh2_instructions[opcode].instrformat,
				i_inst_0_7(opcode),
				i_inst_8_11(opcode));
			break;
		default:
			sprintf(string, "There is an error with sh2_format_instruction_string(%04X)",
				opcode);
			break;
	}
	return;
}

void sh2_dump_instruction(sh2_processor *sh2, uint32_t adr)
{
	uint16_t inst = sh2_read_word(sh2, adr);
	char opstr[512];
	sh2_format_instruction_string(inst, opstr, adr);
	printf("%08X %04X: %s\n", adr, inst, opstr);
}
#endif /* ENABLE_DEBUG */

/* *********************************OPCODES FOLLOW !@#%$# ********************/
void sh2_opcode_nop(sh2_processor *sh2)
{
	sh2->PC += 2;
}

void sh2_opcode_add(sh2_processor *sh2)
{
	sh2->R[i_inst_8_11(sh2->inst)] += sh2->R[i_inst_4_7(sh2->inst)];
	sh2->PC += 2;
}

void sh2_opcode_addi(sh2_processor *sh2)
{
	uint8_t i = i_inst_0_7(sh2->inst);
      
        if ((i & 0x80) == 0)
                sh2->R[i_inst_8_11(sh2->inst)] += i;
        else
                sh2->R[i_inst_8_11(sh2->inst)] += (0xFFFFFF00 | i);
	sh2->PC += 2;
}

void sh2_opcode_addc(sh2_processor *sh2)
{
        uint32_t tmp0, tmp1;

        tmp1 = sh2->R[i_inst_8_11(sh2->inst)] + sh2->R[i_inst_4_7(sh2->inst)];
        tmp0 = sh2->R[i_inst_8_11(sh2->inst)];

        sh2->R[i_inst_8_11(sh2->inst)] = tmp1 + i_get_T(sh2);

        if (tmp0 > tmp1)
                i_set_T(sh2, 1);
        else
                i_set_T(sh2, 0);

        if (tmp1 > sh2->R[i_inst_8_11(sh2->inst)])
                i_set_T(sh2, 1);
        sh2->PC += 2;
}

void sh2_opcode_addv(sh2_processor *sh2)
{
	int dest, src, ans;

	if(sh2->R[i_inst_8_11(sh2->inst)] >= 0)
        	dest = 0;
	else
        	dest = 1;
	
	if(sh2->R[i_inst_4_7(sh2->inst)] >= 0)
        	src = 0;
	else
        	src = 1;
	
	src += dest;
	
	sh2->R[i_inst_8_11(sh2->inst)] += sh2->R[i_inst_4_7(sh2->inst)];
	
	if(sh2->R[i_inst_8_11(sh2->inst)] >= 0)
		ans = 0;
	else
		ans = 1;

	ans += dest;

	if (src == 0 || src == 2)
		i_set_T(sh2, ans == 1);
	else
		i_set_T(sh2, 0);

	sh2->PC += 2;
}

void sh2_opcode_and(sh2_processor *sh2)
{
	sh2->R[i_inst_8_11(sh2->inst)] &= sh2->R[i_inst_4_7(sh2->inst)];
	sh2->PC += 2;
}

void sh2_opcode_andi(sh2_processor *sh2)
{
	sh2->R[0] &= (0x000000FF & i_inst_0_7(sh2->inst));
	sh2->PC += 2;
}

void sh2_opcode_andm(sh2_processor *sh2)
{
	uint8_t temp;
	
	temp  = sh2_read_byte(sh2, sh2->GBR + sh2->R[0]) & i_inst_0_7(sh2->inst);

	sh2_write_byte(sh2, (sh2->GBR + sh2->R[0]), temp);
	sh2->PC += 2;
}


void sh2_opcode_bf(sh2_processor *sh2)
{
	uint32_t disp = i_inst_0_7(sh2->inst);

        if ((disp & 0x80) != 0)
                disp |= 0xFFFFFF00;
                
        if (!i_get_T(sh2))
                sh2->PC = sh2->PC + (disp << 1) + 4;
        else
                sh2->PC += 2;
        
}	

void sh2_opcode_bt(sh2_processor *sh2)
{
	uint32_t disp = i_inst_0_7(sh2->inst);

	if((disp & 0x80) != 0) 
		disp |= 0xFFFFFF00;

	if(i_get_T(sh2)) 
		sh2->PC = sh2->PC + (disp << 1) + 4;
	else 
		sh2->PC += 2;
}

void sh2_opcode_clrmac(sh2_processor *sh2)
{
	sh2->MACH = sh2->MACL = 0;
	sh2->PC += 2;
}

void sh2_opcode_clrt(sh2_processor *sh2)
{
	i_set_T(sh2, 0);
	sh2->PC += 2;
}

void sh2_opcode_sett(sh2_processor *sh2)
{
	i_set_T(sh2, 1);
	sh2->PC += 2;
}

void sh2_opcode_invalid(sh2_processor *sh2)
{
	printf("Invalid opcode (%04X)!\n", sh2->inst);
	printf("R00..: %08X R01..: %08X R02..: %08X R03..: %08X\n", sh2->R[0], sh2->R[1], sh2->R[2], sh2->R[3]);
	printf("R04..: %08X R05..: %08X R06..: %08X R07..: %08X\n", sh2->R[4], sh2->R[5], sh2->R[6], sh2->R[7]);
	printf("R08..: %08X R09..: %08X R10..: %08X R11..: %08X\n", sh2->R[8], sh2->R[9], sh2->R[10], sh2->R[11]);
	printf("R12..: %08X R13..: %08X R14..: %08X R15..: %08X\n", sh2->R[10], sh2->R[11], sh2->R[12], sh2->R[13], sh2->R[14], sh2->R[15]);
	printf("MACH.: %08X MACL.: %08X GBR..: %08X VBR..: %08X\n", sh2->MACH, sh2->MACL, sh2->GBR, sh2->VBR);
	printf("PC...: %08X PR...: %08X SR...: %08X FLAG.: %c%c%c%c\n",
		sh2->PC, sh2->PR, sh2->SR, get_M(sh2) ? 'M' : 'm',
			    get_Q(sh2) ? 'Q' : 'q',
			    get_S(sh2) ? 'S' : 's',
			    get_T(sh2) ? 'T' : 't');

	sh2->invalid_opcode_occ = 1;
}

void sh2_opcode_cmpeq(sh2_processor *sh2)
{
	i_set_T(sh2, sh2->R[i_inst_8_11(sh2->inst)] == sh2->R[i_inst_4_7(sh2->inst)]);
	sh2->PC += 2;
}

void sh2_opcode_cmpge(sh2_processor *sh2)
{
	i_set_T(sh2, sh2->R[i_inst_8_11(sh2->inst)] >= 
        	     sh2->R[i_inst_4_7(sh2->inst)]);
	sh2->PC += 2;
}
	
void sh2_opcode_cmpgt(sh2_processor *sh2)
{
	i_set_T(sh2, sh2->R[i_inst_8_11(sh2->inst)] > 
        	     sh2->R[i_inst_4_7(sh2->inst)]);
	sh2->PC += 2;
}

void sh2_opcode_cmphi(sh2_processor *sh2)
{
	i_set_T(sh2, sh2->R[i_inst_8_11(sh2->inst)] >
		     sh2->R[i_inst_4_7(sh2->inst)]);
	sh2->PC += 2;
}

void sh2_opcode_cmphs(sh2_processor *sh2)
{
	i_set_T(sh2, sh2->R[i_inst_8_11(sh2->inst)] >=
		     sh2->R[i_inst_4_7(sh2->inst)]);
	sh2->PC += 2;
}

void sh2_opcode_cmppl(sh2_processor *sh2)
{
	i_set_T(sh2, sh2->R[i_inst_8_11(sh2->inst)] > 0);
	sh2->PC += 2;
}

void sh2_opcode_cmppz(sh2_processor *sh2)
{
	i_set_T(sh2, sh2->R[i_inst_8_11(sh2->inst)] >= 0);
	sh2->PC += 2;
}

void sh2_opcode_cmpstr(sh2_processor *sh2)
{
	uint32_t temp;
	uint8_t HH, HL, LH, LL;

	temp = sh2->R[i_inst_8_11(sh2->inst)] ^ sh2->R[i_inst_4_7(sh2->inst)];
	HH = (temp >> 12) & 0x000000FF;
	HL = (temp >>  8) & 0x000000FF;
	LH = (temp >>  4) & 0x000000FF;
	LL = temp         & 0x000000FF;

	i_set_T(sh2, !(HH && HL && LH && LL));
	sh2->PC += 2;
}

void sh2_opcode_cmpim(sh2_processor *sh2)
{
	uint32_t imm = i_inst_0_7(sh2->inst);

	if((imm & 0x80) != 0)
		imm |= 0xFFFFFF00;
	
	i_set_T(sh2, sh2->R[0] == imm);
	sh2->PC += 2;
}

void sh2_opcode_div0s(sh2_processor *sh2)
{
	i_set_Q(sh2, sh2->R[i_inst_8_11(sh2->inst)] & 0x80000000);
	i_set_M(sh2, sh2->R[i_inst_4_7(sh2->inst)] & 0x80000000);
	i_set_T(sh2, i_get_Q(sh2)^i_get_M(sh2));
	sh2->PC += 2;
}

void sh2_opcode_div0u(sh2_processor *sh2)
{
	sh2->SR &= ~0x0301;
	sh2->PC += 2;
}

void sh2_opcode_div1(sh2_processor *sh2)
{
	uint32_t tmp0;
	int sign;
	int n = i_inst_8_11(sh2->inst);
	int m = i_inst_4_7(sh2->inst);

	sign = (0x80000000 & sh2->R[n]) != 0;
	
	sh2->R[n] <<= 1;
	sh2->R[n] |= i_get_T(sh2);

	switch(i_get_Q(sh2)) {
		case 0:
			switch(i_get_M(sh2)) {
				case 0:
					tmp0  = sh2->R[n];
					sh2->R[n] -= sh2->R[m];
					i_set_Q(sh2, (sh2->R[n] > tmp0) ^ sign);
					break;
				case 1:
					tmp0  = sh2->R[n];
					sh2->R[n] += sh2->R[m];
					i_set_Q(sh2, (sh2->R[n] >= tmp0) ^ sign);
					break;
			}
			break;
		case 1:
			switch(i_get_M(sh2)) {
				case 0:
					tmp0  = sh2->R[n];
					sh2->R[n] += sh2->R[m];
					i_set_Q(sh2, (sh2->R[n] < tmp0) ^ sign);
					break;
				case 1:
					tmp0  = sh2->R[n];
					sh2->R[n] -= sh2->R[m];
					i_set_Q(sh2, (sh2->R[n] <= tmp0) ^ sign);
					break;
			}
			break;
	}
	i_set_T(sh2, i_get_Q(sh2)^i_get_M(sh2));
	sh2->PC += 2;
}

void sh2_opcode_dmuls(sh2_processor *sh2)
{
	int n = i_inst_8_11(sh2->inst);
	int m = i_inst_4_7(sh2->inst);
	
	uint32_t rnl, rnh, rml, rmh, res0, res1, res2;
	uint32_t temp0, temp1, temp2, temp3;
	int32_t tempm, tempn, fnlml;

	tempn = sh2->R[n];
	tempm = sh2->R[m];

	if(tempn < 0) 
		tempn = 0 - tempn;
	if(tempm < 0)
		tempm = 0 - tempm;
	if((sh2->R[n] ^ sh2->R[m]) < 0)
		fnlml = -1;
	else
		fnlml = 0;

	temp1 = (unsigned long)tempn;
	temp2 = (unsigned long)tempm;

	rnl = temp1 & 0x0000FFFF;
	rnh = (temp1 >> 16) & 0x0000FFFF;
	rml = temp2 & 0x0000FFFF;
	rmh = (temp2 >> 16) & 0x0000FFFF;

	temp0 = rml * rnl;
	temp1 = rmh * rnl;
	temp2 = rml * rnh;
	temp3 = rmh * rnh;

	res2 = 0;
	res1 = temp1 + temp2;
	
	if(res1 < temp1)
		res2 += 0x00010000;
	
	temp1 = (res1 << 16) & 0xFFFF0000;
	res0 = temp0 + temp1;
	if(res0 < temp0)
		res2++;

	res2 = res2 + ((res1 >> 16) & 0x0000FFFF) + temp3;

	if(fnlml < 0) {
		res2 = ~res2;
		if(res0 == 0)
			res2++;
		else
			res0 = (~res0) + 1;
	}
	sh2->MACH = res2;
	sh2->MACL = res0;
	sh2->PC += 2;
}

void sh2_opcode_dmulu(sh2_processor *sh2)
{
	int m = i_inst_4_7(sh2->inst);
	int n = i_inst_8_11(sh2->inst);
	uint32_t rnl, rnh, rml, rmh, res0, res1, res2;
	uint32_t temp0, temp1, temp2, temp3;

	rnl = sh2->R[n] & 0x0000FFFF;
	rnh = (sh2->R[n] >> 16) & 0x0000FFFF;
	rml = sh2->R[m] & 0x0000FFFF;
	rmh = (sh2->R[m] >> 16) & 0x0000FFFF;

	temp0 = rml * rnl;
	temp1 = rmh * rnl;
	temp2 = rml * rnh;
	temp3 = rmh * rnh;
	
	res2 = 0;
	res1 = temp1 + temp2;
	
	if(res1 < temp1)
		res2 += 0x00010000;
	
	temp1 = (res1 << 16) & 0xFFFF0000;
	res0 = temp0 + temp1;

	if(res0 < temp0) 
		res2++;

	res2 = res2 + ((res1 >> 16) & 0x0000FFFF) + temp3;

	sh2->MACH = res2;
	sh2->MACL = res0;
	sh2->PC += 2;
}

void sh2_opcode_dt(sh2_processor *sh2)
{
	sh2->R[i_inst_8_11(sh2->inst)]--;
	i_set_T(sh2, sh2->R[i_inst_8_11(sh2->inst)] == 0);
	sh2->PC += 2;
}

void sh2_opcode_extsb(sh2_processor *sh2)
{
	sh2->R[i_inst_8_11(sh2->inst)] = sh2->R[i_inst_4_7(sh2->inst)];

	if((sh2->R[i_inst_4_7(sh2->inst)] & 0x00000080) == 0)
		sh2->R[i_inst_8_11(sh2->inst)] &= 0x000000FF;
	else
		sh2->R[i_inst_8_11(sh2->inst)] |= 0xFFFFFF00;
	sh2->PC += 2;
}

void sh2_opcode_extsw(sh2_processor *sh2)
{
	sh2->R[i_inst_8_11(sh2->inst)] = sh2->R[i_inst_4_7(sh2->inst)];

	if((sh2->R[i_inst_4_7(sh2->inst)] & 0x00008000) == 0)
		sh2->R[i_inst_8_11(sh2->inst)] &= 0x0000FFFF;
	else
		sh2->R[i_inst_4_7(sh2->inst)] |= 0xFFFF0000;
	sh2->PC += 2;
}
	
void sh2_opcode_extub(sh2_processor *sh2)
{
	sh2->R[i_inst_8_11(sh2->inst)]  = sh2->R[i_inst_4_7(sh2->inst)];
	sh2->R[i_inst_8_11(sh2->inst)] &= 0x000000FF;
	sh2->PC += 2;
}

void sh2_opcode_extuw(sh2_processor *sh2)
{
	sh2->R[i_inst_8_11(sh2->inst)]  = sh2->R[i_inst_4_7(sh2->inst)];
	sh2->R[i_inst_8_11(sh2->inst)] &= 0x0000FFFF;
	sh2->PC += 2;
}

void sh2_opcode_sleep(sh2_processor *sh2)
{
	/* Basically a NOP w/o the PC increment */
	sh2->processor_sleep = 1;
	sh2->PC = sh2->PC;
}

void sh2_opcode_bfs(sh2_processor *sh2)
{
	uint32_t disp = i_inst_0_7(sh2->inst);
	uint32_t temp;

	temp = sh2->PC;

	if((disp & 0x80) != 0) 
		disp |= 0xFFFFFF00;
	if(!i_get_T(sh2)) {
		sh2->PC = sh2->PC + (disp << 1) + 4;
		delay_slot(sh2, temp + 2);
	} else 
		sh2->PC += 2;
}

void sh2_opcode_bts(sh2_processor *sh2)
{
	uint32_t disp = i_inst_0_7(sh2->inst);
	uint32_t temp;

	temp = sh2->PC;
	if((disp & 0x80) != 0) 
		disp |= 0xFFFFFF00;
	if(i_get_T(sh2)) {
		sh2->PC = sh2->PC + (disp << 1) + 4;
		delay_slot(sh2, temp + 2);
        }
        else 
		sh2->PC += 2;
}

void sh2_opcode_bra(sh2_processor *sh2)
{
	uint32_t temp;
	uint32_t disp = i_inst_0_11(sh2->inst);
	
	if((disp & 0x800) != 0) 
		disp |= 0xFFFFF000;
	temp = sh2->PC;
	sh2->PC = sh2->PC + (disp << 1) + 4;

	delay_slot(sh2, temp+2);

}

void sh2_opcode_braf(sh2_processor *sh2)
{
	uint32_t temp;

        temp = sh2->PC;
        sh2->PC  += sh2->R[i_inst_8_11(sh2->inst)] + 4;

        delay_slot(sh2, temp + 2);

}

void sh2_opcode_bsr(sh2_processor *sh2)
{
	uint32_t disp = i_inst_0_11(sh2->inst);

        if((disp & 0x800) != 0) 
		disp |= 0xFFFFF000;
        sh2->PR = sh2->PC + 4;
        sh2->PC = sh2->PC + (disp << 1) + 4;
        delay_slot(sh2, sh2->PR - 2);
}

void sh2_opcode_bsrf(sh2_processor *sh2)
{
	sh2->PR  = sh2->PC + 4;
	sh2->PC += sh2->R[i_inst_8_11(sh2->inst)] + 4;
	delay_slot(sh2, sh2->PR - 2);
}

void sh2_opcode_jmp(sh2_processor *sh2)
{
	uint32_t temp;
	
	temp = sh2->PC;
	sh2->PC   = sh2->R[i_inst_8_11(sh2->inst)];
	delay_slot(sh2, temp + 2);
}

void sh2_opcode_jsr(sh2_processor *sh2)
{
	sh2->PR = sh2->PC + 4;
	sh2->PC = sh2->R[i_inst_8_11(sh2->inst)];
	delay_slot(sh2, sh2->PR - 2);
}

void sh2_opcode_ldcsr(sh2_processor *sh2)
{
	sh2->SR  = sh2->R[i_inst_8_11(sh2->inst)] & 0x000003F3;
	sh2->PC += 2;
}

void sh2_opcode_ldcgbr(sh2_processor *sh2)
{
	sh2->GBR = sh2->R[i_inst_8_11(sh2->inst)];
	sh2->PC += 2;
}

void sh2_opcode_ldcvbr(sh2_processor *sh2)
{
	sh2->VBR = sh2->R[i_inst_8_11(sh2->inst)];
	sh2->PC += 2;
}

void sh2_opcode_ldcmsr(sh2_processor *sh2)
{
	sh2->SR  = sh2_read_long(sh2, sh2->R[i_inst_8_11(sh2->inst)]) & 0x000003F3;
	sh2->R[i_inst_8_11(sh2->inst)] += 4;
	sh2->PC += 2;
}

void sh2_opcode_ldcmgbr(sh2_processor *sh2)
{
	sh2->GBR = sh2_read_long(sh2, sh2->R[i_inst_8_11(sh2->inst)]);
	sh2->R[i_inst_8_11(sh2->inst)] += 4;
	sh2->PC += 2;
}

void sh2_opcode_ldcmvbr(sh2_processor *sh2)
{
	sh2->VBR = sh2_read_long(sh2, sh2->R[i_inst_8_11(sh2->inst)]);
	sh2->R[i_inst_8_11(sh2->inst)] += 4;
	sh2->PC += 2;
}

	
void sh2_opcode_ldsmach(sh2_processor *sh2)
{
	sh2->MACH = sh2->R[i_inst_8_11(sh2->inst)];
	sh2->PC += 2;
}

void sh2_opcode_ldsmacl(sh2_processor *sh2)
{
	sh2->MACL = sh2->R[i_inst_8_11(sh2->inst)];
	sh2->PC  += 2;
}

void sh2_opcode_ldspr(sh2_processor *sh2)
{
	sh2->PR = sh2->R[i_inst_8_11(sh2->inst)];
	sh2->PC += 2;
}

void sh2_opcode_ldsmmach(sh2_processor *sh2)
{
	sh2->MACH = sh2_read_long(sh2, sh2->R[i_inst_8_11(sh2->inst)]);
	sh2->R[i_inst_8_11(sh2->inst)] += 4;
	sh2->PC += 2;
}

void sh2_opcode_ldsmmacl(sh2_processor *sh2)
{
	sh2->MACL = sh2_read_long(sh2, sh2->R[i_inst_8_11(sh2->inst)]);
	sh2->R[i_inst_8_11(sh2->inst)] += 4;
	sh2->PC += 2;
}

void sh2_opcode_ldsmpr(sh2_processor *sh2)
{
	sh2->PR = sh2_read_long(sh2, sh2->R[i_inst_8_11(sh2->inst)]);
	sh2->R[i_inst_8_11(sh2->inst)] += 4;
	sh2->PC += 2;
}

void sh2_opcode_macl(sh2_processor *sh2)
{
	int m = i_inst_4_7(sh2->inst);
	int n = i_inst_8_11(sh2->inst);
	uint32_t rnl, rnh, rml, rmh, res0, res1, res2;
	uint32_t temp0, temp1, temp2, temp3;
	int32_t tempm, tempn, fnlml;

	tempn = sh2_read_long(sh2, sh2->R[n]);
	sh2->R[n] += 4;
	tempm = sh2_read_long(sh2, sh2->R[m]);
	sh2->R[m] += 4;

	if((tempn ^ tempm) < 0)
		fnlml = -1;
	else 
		fnlml = 0;

	if(tempn < 0) 
		tempn = 0 - tempn;
        if(tempm < 0) 
		tempm = 0 - tempm;

	temp1 = tempn;
	temp2 = tempm;

	rnl   = temp1 & 0x0000FFFF;
	rnh   = (temp1 >> 16) & 0x0000FFFF;
	rml   = temp2 & 0x0000FFFF;
	rmh   = (temp2 >> 16) & 0x0000FFFF;

	temp0 = rml*rnl;
	temp1 = rmh*rnl;
	temp2 = rml*rnh;
	temp3 = rmh*rnh;

	res2  = 0;
	res1  = temp1 + temp2;
	if(res1 < temp1) 
		res2 += 0x00010000;

	temp1 = (res1 << 16) & 0xFFFF0000;
	res0  = temp0 + temp1;
	if(res0 < temp0) 
		res2++;

	res2  = res2 + ((res1 >> 16) & 0x0000FFFF) + temp3;

	if(fnlml < 0) {
		res2 = ~res2;
		if(res0 == 0) 
			res2++;
		else 
			res0 = (~res0) + 1;
	}
	if(i_get_S(sh2)) {
		res0 = sh2->MACL + res0;
		if(sh2->MACL > res0) 
			res2++;
		res2 += (sh2->MACH & 0x0000FFFF);

		if((res2 < 0) && (res2 < 0xFFFF8000)) {
			res2 = 0x00008000;
			res0 = 0x00000000;
		}
		if((res2 > 0) && (res2 > 0x00007FFF)) {
			res2 = 0x00007FFF;
			res0 = 0xFFFFFFFF;
		};
		sh2->MACH = res2;
		sh2->MACL = res0;
	} else {
		res0 = sh2->MACL + res0;
		if(sh2->MACL > res0) 
			res2++;

		res2 += sh2->MACH;
		sh2->MACH = res2;
		sh2->MACL = res0;
	}
	sh2->PC += 2;
}
 
void sh2_opcode_macw(sh2_processor *sh2)
{
	int32_t tempm, tempn, dest, src, ans;
	uint32_t templ;
	int n = i_inst_8_11(sh2->inst);
	int m = i_inst_4_7(sh2->inst);

	tempn = sh2_read_word(sh2, sh2->R[n]);
	sh2->R[n] += 2;
	tempm = sh2_read_word(sh2, sh2->R[m]);
	sh2->R[m] += 2;
	templ = sh2->MACL;
	tempm = ((int16_t)tempn * (int16_t)tempm);

	if(sh2->MACL >= 0) 
		dest = 0;
	else 
		dest = 1;
	if(tempm >= 0) {
		src = 0;
		tempn = 0;
	} else {
		src = 1;
		tempn = 0xFFFFFFFF;
	}

	src  += dest;
	sh2->MACL += tempm;
	if(sh2->MACL >= 0) 
		ans = 0;
	else 
		ans = 1;
	ans  += dest;

	if(i_get_S(sh2)) {
		if(ans == 1) {
			if(src == 0) 
				sh2->MACL = 0x7FFFFFFF;
			if(src == 2) 
				sh2->MACL = 0x80000000;
		}
	} else {
		sh2->MACH += tempn;
		if(templ > sh2->MACL) 
			sh2->MACH += 1;
	}
	sh2->PC += 2;
}

void sh2_opcode_movbs0(sh2_processor *sh2)
{
	sh2_write_byte(sh2, sh2->R[i_inst_8_11(sh2->inst)] + sh2->R[0],
		       sh2->R[i_inst_4_7(sh2->inst)]);
	sh2->PC += 2;
}

void sh2_opcode_movws0(sh2_processor *sh2)
{
	sh2_write_word(sh2, sh2->R[i_inst_8_11(sh2->inst)] + sh2->R[0],
		       sh2->R[i_inst_4_7(sh2->inst)]);
	sh2->PC += 2;
}
	                                          
void sh2_opcode_movls0(sh2_processor *sh2)
{
        sh2_write_long(sh2, sh2->R[i_inst_8_11(sh2->inst)] + sh2->R[0],
                       sh2->R[i_inst_4_7(sh2->inst)]);
        sh2->PC += 2;
}	
         
void sh2_opcode_mov(sh2_processor *sh2)
{
	sh2->R[i_inst_8_11(sh2->inst)] = sh2->R[i_inst_4_7(sh2->inst)];
	sh2->PC += 2;
}

void sh2_opcode_movbs(sh2_processor *sh2)
{
	sh2_write_byte(sh2, sh2->R[i_inst_8_11(sh2->inst)], sh2->R[i_inst_4_7(sh2->inst)]);
	sh2->PC += 2;
}

void sh2_opcode_movws(sh2_processor *sh2)
{
	sh2_write_word(sh2, sh2->R[i_inst_8_11(sh2->inst)], sh2->R[i_inst_4_7(sh2->inst)]);
	sh2->PC += 2;
}

void sh2_opcode_movls(sh2_processor *sh2)
{
	sh2_write_long(sh2, sh2->R[i_inst_8_11(sh2->inst)], sh2->R[i_inst_4_7(sh2->inst)]);
	sh2->PC += 2;
}

void sh2_opcode_movbl(sh2_processor *sh2)
{
	sh2->R[i_inst_8_11(sh2->inst)] = sh2_read_byte(sh2, sh2->R[i_inst_4_7(sh2->inst)]);
	if((sh2->R[i_inst_8_11(sh2->inst)] & 0x80) != 0) 
		sh2->R[i_inst_8_11(sh2->inst)] |= 0xFFFFFF00;
	sh2->PC += 2;
}

void sh2_opcode_movwl(sh2_processor *sh2)
{
	sh2->R[i_inst_8_11(sh2->inst)] = sh2_read_word(sh2, sh2->R[i_inst_4_7(sh2->inst)]);
	if((sh2->R[i_inst_8_11(sh2->inst)] & 0x8000) != 0)
		sh2->R[i_inst_8_11(sh2->inst)] |= 0xFFFF0000;
	sh2->PC += 2;
}

void sh2_opcode_movll(sh2_processor *sh2)
{
	sh2->R[i_inst_8_11(sh2->inst)] = sh2_read_long(sh2, sh2->R[i_inst_4_7(sh2->inst)]);
	sh2->PC += 2;
}

void sh2_opcode_movbm(sh2_processor *sh2)
{
	sh2_write_byte(sh2, sh2->R[i_inst_8_11(sh2->inst)] - 1, 
		       sh2->R[i_inst_4_7(sh2->inst)]);
	sh2->R[i_inst_8_11(sh2->inst)] -= 1;
	sh2->PC += 2;
}

void sh2_opcode_movwm(sh2_processor *sh2)
{
	sh2_write_word(sh2, sh2->R[i_inst_8_11(sh2->inst)] - 2,
		       sh2->R[i_inst_4_7(sh2->inst)]);
	sh2->R[i_inst_8_11(sh2->inst)] -= 2;
	sh2->PC += 2;
}

void sh2_opcode_movlm(sh2_processor *sh2)
{
	sh2_write_long(sh2, sh2->R[i_inst_8_11(sh2->inst)] - 4,
		       sh2->R[i_inst_4_7(sh2->inst)]);
	sh2->R[i_inst_8_11(sh2->inst)] -= 4;
	sh2->PC += 2;
}

void sh2_opcode_movi(sh2_processor *sh2)
{
	uint8_t i = i_inst_0_7(sh2->inst);
	int n = i_inst_8_11(sh2->inst);
	
	if((i & 0x80) == 0) 
		sh2->R[n] = i;
	else 
		sh2->R[n] = (0xFFFFFF00 | i);
	sh2->PC += 2;
}

void sh2_opcode_movwi(sh2_processor *sh2)
{
	uint32_t disp = i_inst_0_7(sh2->inst);
	int n = i_inst_8_11(sh2->inst);
	
	sh2->R[n] = sh2_read_word(sh2, (sh2->PC+4) + (disp << 1));
	if((sh2->R[n] & 0x8000) == 0)
		sh2->R[n] &= 0x0000FFFF;
	else 
		sh2->R[n] |= 0xFFFF0000;
	sh2->PC += 2;
}	                              

void sh2_opcode_movli(sh2_processor *sh2)
{
	uint32_t disp = i_inst_0_7(sh2->inst);
	int n = i_inst_8_11(sh2->inst);

	sh2->R[n] = sh2_read_long(sh2, ((sh2->PC+4) & 0xFFFFFFFC) + (disp << 2));
	sh2->PC += 2;
}

void sh2_opcode_movbp(sh2_processor *sh2)
{
	int n = i_inst_8_11(sh2->inst);
	int m = i_inst_4_7(sh2->inst);
	sh2->R[n] = sh2_read_byte(sh2, sh2->R[m]);
	if((sh2->R[n] & 0x80) != 0) 
		sh2->R[n] |= 0xFFFFFF00;
	if(n != m)
		sh2->R[m] += 1;
	sh2->PC += 2;
}

void sh2_opcode_movwp(sh2_processor *sh2)
{
	int n = i_inst_8_11(sh2->inst);
	int m = i_inst_4_7(sh2->inst);
                
	sh2->R[n] = sh2_read_word(sh2, sh2->R[m]);
	if((sh2->R[n] & 0x8000) != 0) 
		sh2->R[n] |= 0xFFFF0000;
	if(n != m) 
		sh2->R[m] += 2;
	sh2->PC+=2;
}

void sh2_opcode_movlp(sh2_processor *sh2)
{
	int n = i_inst_8_11(sh2->inst);
	int m = i_inst_4_7(sh2->inst);
               
	sh2->R[n] = sh2_read_long(sh2, sh2->R[m]);
	if(n != m)
		sh2->R[m] += 4;
	sh2->PC += 2;
}

void sh2_opcode_movbl0(sh2_processor *sh2)
{
	int n = i_inst_8_11(sh2->inst);
	int m = i_inst_4_7(sh2->inst);

	sh2->R[n] = sh2_read_byte(sh2, sh2->R[m] + sh2->R[0]);
	if((sh2->R[n] & 0x80) != 0) 
		sh2->R[n] |= 0xFFFFFF00;
	sh2->PC += 2;
}

void sh2_opcode_movwl0(sh2_processor *sh2)
{
	int n = i_inst_8_11(sh2->inst);
	int m = i_inst_4_7(sh2->inst);

	sh2->R[n] = sh2_read_word(sh2, sh2->R[m] + sh2->R[0]);
	if((sh2->R[n] & 0x8000) != 0) 
		sh2->R[n] |= 0xFFFF0000;
	sh2->PC += 2;
}

void sh2_opcode_movll0(sh2_processor *sh2)
{
	int n = i_inst_8_11(sh2->inst);
	int m = i_inst_4_7(sh2->inst);
                
	sh2->R[n] = sh2_read_long(sh2, sh2->R[m] + sh2->R[0]);
	sh2->PC += 2;
}

void sh2_opcode_movblg(sh2_processor *sh2)
{
	uint32_t disp = i_inst_0_7(sh2->inst);
      
	sh2->R[0] = sh2_read_byte(sh2, sh2->GBR + disp);
	if((sh2->R[0] & 0x80) != 0) 
		sh2->R[0] |= 0xFFFFFF00;
	sh2->PC += 2;
}

void sh2_opcode_movwlg(sh2_processor *sh2)
{
	uint32_t disp = i_inst_0_7(sh2->inst);      
      
	sh2->R[0] = sh2_read_word(sh2, sh2->GBR + (disp << 1));
	if((sh2->R[0] & 0x8000) != 0) 
		sh2->R[0] |= 0xFFFF0000;
	sh2->PC += 2;
}

void sh2_opcode_movllg(sh2_processor *sh2)
{
	uint32_t disp = i_inst_0_7(sh2->inst);

	sh2->R[0] = sh2_read_long(sh2, sh2->GBR + (disp << 2));
	sh2->PC += 2;
}
                  	                                                                        
void sh2_opcode_movbsg(sh2_processor *sh2)
{
	uint32_t disp = i_inst_0_7(sh2->inst);
	
	sh2_write_byte(sh2, sh2->GBR + disp, sh2->R[0]);
	sh2->PC += 2;
}
	                  
void sh2_opcode_movwsg(sh2_processor *sh2)
{
	uint32_t disp = i_inst_0_7(sh2->inst);
	
	sh2_write_word(sh2, sh2->GBR + (disp << 1), sh2->R[0]);
	sh2->PC += 2;
}

void sh2_opcode_movlsg(sh2_processor *sh2)
{
	uint32_t disp = i_inst_0_7(sh2->inst);
	
	sh2_write_long(sh2, sh2->GBR+(disp<<2),sh2->R[0]);
	sh2->PC += 2;
}
                  	
void sh2_opcode_movbs4(sh2_processor *sh2)
{
	uint32_t disp = i_inst_0_3(sh2->inst);
	int n = i_inst_4_7(sh2->inst);
      
	sh2_write_byte(sh2, sh2->R[n] + disp, sh2->R[0]);
	sh2->PC += 2;
}
                        	
void sh2_opcode_movws4(sh2_processor *sh2)
{
	uint32_t disp = i_inst_0_3(sh2->inst);
	int n = i_inst_4_7(sh2->inst);

	sh2_write_word(sh2, sh2->R[n] + (disp << 1), sh2->R[0]);
	sh2->PC += 2;
}
                                          
void sh2_opcode_movls4(sh2_processor *sh2)
{
	uint32_t disp = i_inst_0_3(sh2->inst);
	int n = i_inst_8_11(sh2->inst);
	int m = i_inst_4_7(sh2->inst);
	
	sh2_write_long(sh2, sh2->R[n] + (disp << 2), sh2->R[m]);
	sh2->PC += 2;
}

void sh2_opcode_movbl4(sh2_processor *sh2)
{
	uint32_t disp = i_inst_0_3(sh2->inst);
	int m = i_inst_4_7(sh2->inst);

	sh2->R[0] = sh2_read_byte(sh2, sh2->R[m] + disp);
	if((sh2->R[0] & 0x80) != 0) 
		sh2->R[0] |= 0xFFFFFF00;
	sh2->PC += 2;
}

void sh2_opcode_movwl4(sh2_processor *sh2)
{
	uint32_t disp = i_inst_0_3(sh2->inst);
	int m = i_inst_4_7(sh2->inst);

	sh2->R[0] = sh2_read_word(sh2, sh2->R[m] + (disp << 1));
	if((sh2->R[0] & 0x8000) != 0) 
		sh2->R[0] |= 0xFFFF0000;
	sh2->PC += 2;
}

void sh2_opcode_movll4(sh2_processor *sh2)
{
        uint32_t disp = i_inst_0_3(sh2->inst);
        int n = i_inst_8_11(sh2->inst);
        int m = i_inst_4_7(sh2->inst);

	sh2->R[n] = sh2_read_long(sh2, sh2->R[m] + (disp << 2));
	sh2->PC += 2;
}

void sh2_opcode_mova(sh2_processor *sh2)
{
	uint32_t disp;
	
	disp = i_inst_0_7(sh2->inst);
	sh2->R[0] = (sh2->PC & 0xFFFFFFFC) + (disp << 2) + 4;
	sh2->PC += 2;
}

void sh2_opcode_movt(sh2_processor *sh2)
{
	sh2->R[i_inst_8_11(sh2->inst)] = i_get_T(sh2);
	sh2->PC += 2;
}

void sh2_opcode_mull(sh2_processor *sh2)
{
	sh2->MACL = sh2->R[i_inst_8_11(sh2->inst)] * sh2->R[i_inst_4_7(sh2->inst)];
	sh2->PC += 2;
}

void sh2_opcode_muls(sh2_processor *sh2)
{
	sh2->MACL = ((int16_t)sh2->R[i_inst_8_11(sh2->inst)] * (int16_t)sh2->R[i_inst_4_7(sh2->inst)]);
	sh2->PC += 2;
}

void sh2_opcode_mulu(sh2_processor *sh2)
{
	sh2->MACL = ((uint32_t)(uint16_t)sh2->R[i_inst_8_11(sh2->inst)]
		* (uint32_t)(uint16_t)sh2->R[i_inst_4_7(sh2->inst)]);
	sh2->PC += 2;
}

void sh2_opcode_neg(sh2_processor *sh2)
{
	sh2->R[i_inst_8_11(sh2->inst)] = 0 - sh2->R[i_inst_4_7(sh2->inst)];
	sh2->PC += 2;
}

void sh2_opcode_negc(sh2_processor *sh2)
{
	int32_t temp;
	
	temp = 0 - sh2->R[i_inst_4_7(sh2->inst)];
	sh2->R[i_inst_8_11(sh2->inst)] = temp - i_get_T(sh2);
	i_set_T(sh2, (0 < temp) || (temp < sh2->R[i_inst_8_11(sh2->inst)]));
	sh2->PC += 2;
}

void sh2_opcode_not(sh2_processor *sh2)
{
	sh2->R[i_inst_8_11(sh2->inst)] = ~sh2->R[i_inst_4_7(sh2->inst)];
	sh2->PC += 2;
}

void sh2_opcode_or(sh2_processor *sh2)
{
	sh2->R[i_inst_8_11(sh2->inst)] |= sh2->R[i_inst_4_7(sh2->inst)];
	sh2->PC += 2;
}

void sh2_opcode_ori(sh2_processor *sh2)
{
	sh2->R[0] |= i_inst_0_7(sh2->inst);
	sh2->PC += 2;
}

void sh2_opcode_orm(sh2_processor *sh2)
{
	uint8_t temp;
	
	temp  = sh2_read_byte(sh2, sh2->GBR + sh2->R[0]);
	temp |= i_inst_0_7(sh2->inst);
	sh2_write_byte(sh2, sh2->GBR + sh2->R[0], temp);
	sh2->PC += 2;
}

void sh2_opcode_rotcl(sh2_processor *sh2)
{
	int temp;
	int n = i_inst_8_11(sh2->inst);
	
	temp = (sh2->R[n] & 0x80000000) != 0;

	sh2->R[n] <<= 1;

	if(i_get_T(sh2)) 
		sh2->R[n] |= 0x00000001;
	else 
		sh2->R[n] &= 0xFFFFFFFE;

        i_set_T(sh2, temp);
	sh2->PC += 2;
}

void sh2_opcode_rotcr(sh2_processor *sh2)
{
	int temp;
	int n = i_inst_8_11(sh2->inst);
	
	temp = (sh2->R[n] & 0x00000001) != 0;

	sh2->R[n] >>= 1;
	if(i_get_T(sh2)) 
		sh2->R[n] |= 0x80000000;
	else 
		sh2->R[n] &= 0x7FFFFFFF;
	i_set_T(sh2, temp);
	sh2->PC += 2;
}

void sh2_opcode_rotl(sh2_processor *sh2)
{
	int n = i_inst_8_11(sh2->inst);
	i_set_T(sh2, sh2->R[n] & 0x80000000);

	sh2->R[n] <<= 1;
	if(i_get_T(sh2)) 
		sh2->R[n] |= 0x00000001;
	else 
		sh2->R[n] &= 0xFFFFFFFE;

	sh2->PC += 2;
}

void sh2_opcode_rotr(sh2_processor *sh2)
{
	int n = i_inst_8_11(sh2->inst);
	i_set_T(sh2, sh2->R[n] & 0x00000001);

	sh2->R[n] >>= 1;

	if(i_get_T(sh2)) 
		sh2->R[n] |= 0x80000000;
	else 
		sh2->R[n] &= 0x7FFFFFFF;

	sh2->PC += 2;
}

void sh2_opcode_rte(sh2_processor *sh2)
{
	uint32_t temp;
	temp = sh2->PC;
	
	sh2->PC = sh2_read_long(sh2, sh2->R[15]);
	sh2->R[15] += 4;
	
	sh2->SR = sh2_read_long(sh2, sh2->R[15]) & 0x000003F3;
	sh2->R[15] += 4;
	
	delay_slot(sh2, temp + 2);
}

void sh2_opcode_rts(sh2_processor *sh2)
{
	uint32_t temp;
	
	temp = sh2->PC;
	sh2->PC = sh2->PR;

	delay_slot(sh2, temp + 2);
}

void sh2_opcode_shal(sh2_processor *sh2)
{
	i_set_T(sh2, sh2->R[i_inst_8_11(sh2->inst)] & 0x80000000);
		
	sh2->R[i_inst_8_11(sh2->inst)] <<= 1;
	sh2->PC += 2;
}

void sh2_opcode_shar(sh2_processor *sh2)
{
	int temp;
	
	if((sh2->R[i_inst_8_11(sh2->inst)] & 0x00000001) == 0)
		i_set_T(sh2, 0);
	else
		i_set_T(sh2, 1);
	
	temp = (sh2->R[i_inst_8_11(sh2->inst)] & 0x80000000) != 0;
	
	sh2->R[i_inst_8_11(sh2->inst)] >>= 1;
	
	if(temp)
		sh2->R[i_inst_8_11(sh2->inst)] |= 0x80000000;
	else
		sh2->R[i_inst_8_11(sh2->inst)] &= 0x7FFFFFFF;
		
	sh2->PC += 2;
}

void sh2_opcode_shll2(sh2_processor *sh2)
{
	sh2->R[i_inst_8_11(sh2->inst)] <<= 2;
	sh2->PC += 2;
}

void sh2_opcode_shll8(sh2_processor *sh2)
{
	sh2->R[i_inst_8_11(sh2->inst)] <<= 8;
	sh2->PC += 2;
}

void sh2_opcode_shll16(sh2_processor *sh2)
{
	sh2->R[i_inst_8_11(sh2->inst)] <<= 16;
	sh2->PC += 2;
}

void sh2_opcode_shlr(sh2_processor *sh2)
{
	i_set_T(sh2, sh2->R[i_inst_8_11(sh2->inst)] & 0x00000001);
	sh2->R[i_inst_8_11(sh2->inst)] >>= 1;
	sh2->R[i_inst_8_11(sh2->inst)]  &= 0x7FFFFFFF;
	sh2->PC += 2;
}

void sh2_opcode_shlr2(sh2_processor *sh2)
{
	sh2->R[i_inst_8_11(sh2->inst)] >>= 2;
	sh2->R[i_inst_8_11(sh2->inst)]  &= 0x3FFFFFFF;
	sh2->PC += 2;
}

void sh2_opcode_shlr8(sh2_processor *sh2)
{
	sh2->R[i_inst_8_11(sh2->inst)] >>= 8;
	sh2->R[i_inst_8_11(sh2->inst)]  &= 0x00FFFFFF;
	sh2->PC += 2;
}

void sh2_opcode_shlr16(sh2_processor *sh2)
{
	sh2->R[i_inst_8_11(sh2->inst)] >>= 16;
	sh2->R[i_inst_8_11(sh2->inst)]  &= 0x0000FFFF;
	sh2->PC += 2;
}

void sh2_opcode_stcsr(sh2_processor *sh2)
{
	sh2->R[i_inst_8_11(sh2->inst)] = sh2->SR;
	sh2->PC += 2;
}

void sh2_opcode_stcgbr(sh2_processor *sh2)
{
	sh2->R[i_inst_8_11(sh2->inst)] = sh2->GBR;
	sh2->PC += 2;
}

void sh2_opcode_stcvbr(sh2_processor *sh2)
{
	sh2->R[i_inst_8_11(sh2->inst)] = sh2->VBR;
	sh2->PC += 2;
}

void sh2_opcode_stcmsr(sh2_processor *sh2)
{
	sh2->R[i_inst_8_11(sh2->inst)] -= 4;
	sh2_write_long(sh2, sh2->R[i_inst_8_11(sh2->inst)], sh2->SR);
	sh2->PC += 2;
}

void sh2_opcode_stcmgbr(sh2_processor *sh2)
{
	sh2->R[i_inst_8_11(sh2->inst)] -= 4;
	sh2_write_long(sh2, sh2->R[i_inst_8_11(sh2->inst)], sh2->GBR);
	sh2->PC += 2;
}

void sh2_opcode_stcmvbr(sh2_processor *sh2)
{
	sh2->R[i_inst_8_11(sh2->inst)] -= 4;
	sh2_write_long(sh2, sh2->R[i_inst_8_11(sh2->inst)], sh2->VBR);
	sh2->PC += 2;
}

void sh2_opcode_stsmach(sh2_processor *sh2)
{
	sh2->R[i_inst_8_11(sh2->inst)] = sh2->MACH;
	sh2->PC += 2;
}
	
void sh2_opcode_stsmacl(sh2_processor *sh2)
{
	sh2->R[i_inst_8_11(sh2->inst)] = sh2->MACL;
	sh2->PC += 2;
}

void sh2_opcode_stspr(sh2_processor *sh2)
{
	sh2->R[i_inst_8_11(sh2->inst)] = sh2->PR;
	sh2->PC += 2;
}

void sh2_opcode_stsmmach(sh2_processor *sh2)
{
	sh2->R[i_inst_8_11(sh2->inst)] -= 4;
	sh2_write_long(sh2, sh2->R[i_inst_8_11(sh2->inst)], sh2->MACH);
	sh2->PC += 2;
}

void sh2_opcode_stsmmacl(sh2_processor *sh2)
{
	sh2->R[i_inst_8_11(sh2->inst)] -= 4;
	sh2_write_long(sh2, sh2->R[i_inst_8_11(sh2->inst)], sh2->MACL);
	sh2->PC += 2;
}

void sh2_opcode_stsmpr(sh2_processor *sh2)
{
	sh2->R[i_inst_8_11(sh2->inst)] -= 4;
	sh2_write_long(sh2, sh2->R[i_inst_8_11(sh2->inst)], sh2->PR);
	sh2->PC += 2;
}

void sh2_opcode_sub(sh2_processor *sh2)
{
	sh2->R[i_inst_8_11(sh2->inst)] -= sh2->R[i_inst_4_7(sh2->inst)];
	sh2->PC += 2;
}

void sh2_opcode_subc(sh2_processor *sh2)
{
	uint32_t tmp0, tmp1;
	
	tmp1 = sh2->R[i_inst_8_11(sh2->inst)] - sh2->R[i_inst_4_7(sh2->inst)];
	tmp0 = sh2->R[i_inst_8_11(sh2->inst)];
	sh2->R[i_inst_8_11(sh2->inst)] = tmp1 - i_get_T(sh2);
	
	i_set_T(sh2, (tmp0 < tmp1) || (tmp1 < sh2->R[i_inst_8_11(sh2->inst)]));
	
	sh2->PC += 2;
}

void sh2_opcode_subv(sh2_processor *sh2)
{
	int n = i_inst_8_11(sh2->inst);
	int m = i_inst_4_7(sh2->inst);
	int dest, src, ans;
	
	if(sh2->R[n] >= 0)
		dest = 0;
	else
		dest = 1;
	if(sh2->R[m] >= 0)
		src = 0;
	else
		src = 1;
	
	src += dest;
	
	sh2->R[n] -= sh2->R[m];
	
	if(sh2->R[n] >= 0)
		ans = 0;
	else
		ans = 1;
	
	ans += dest;
	
	i_set_T(sh2, (src == 1) && (ans == 1));
	sh2->PC += 2;
}

void sh2_opcode_swapb(sh2_processor *sh2)
{
	int m = i_inst_4_7(sh2->inst);
	int n = i_inst_8_11(sh2->inst);
	
	sh2->R[n] = (sh2->R[m] & 0xFFFF0000) | 
		   ((sh2->R[m] & 0x000000FF) << 8) |
	           ((sh2->R[m] & 0x0000FF00) >> 8);

	sh2->PC += 2;
}

void sh2_opcode_swapw(sh2_processor *sh2)
{
	int m = i_inst_4_7(sh2->inst);
	int n = i_inst_8_11(sh2->inst);

	sh2->R[n] = ((sh2->R[m] & 0xFFFF0000) >> 16) |
		   ((sh2->R[m] & 0x0000FFFF) << 16);

	sh2->PC += 2;
}

void sh2_opcode_tas(sh2_processor *sh2)
{
	uint8_t temp;
	
	temp =  sh2_read_byte(sh2, sh2->R[i_inst_8_11(sh2->inst)]);
	i_set_T(sh2, temp == 0);
	
	temp |= 0x00000080;
	
	sh2_write_byte(sh2, sh2->R[i_inst_8_11(sh2->inst)], temp);
	sh2->PC += 2;
}

void sh2_opcode_trapa(sh2_processor *sh2)
{
	uint32_t imm = i_inst_0_7(sh2->inst);
	
	sh2->R[15] -= 4;
	sh2_write_long(sh2, sh2->R[15], sh2->SR);
	
	sh2->R[15] -= 4;
	sh2_write_long(sh2, sh2->R[15], sh2->PC - 2);
	sh2->PC = sh2_read_long(sh2, sh2->VBR + (imm << 2)) + 4;
}

void sh2_opcode_tst(sh2_processor *sh2)
{

	i_set_T(sh2, ((sh2->R[i_inst_8_11(sh2->inst)] & 
		     sh2->R[i_inst_4_7(sh2->inst)]) == 0));

	sh2->PC += 2;
}

void sh2_opcode_tsti(sh2_processor *sh2)
{
	i_set_T(sh2, (sh2->R[0] &  i_inst_0_7(sh2->inst)) == 0);

	sh2->PC += 2;
}

void sh2_opcode_tstm(sh2_processor *sh2)
{
	i_set_T(sh2, sh2_read_byte(sh2, sh2->GBR + sh2->R[0]) & i_inst_0_7(sh2->inst));

	sh2->PC += 2;
}

void sh2_opcode_xor(sh2_processor *sh2)
{
	sh2->R[i_inst_8_11(sh2->inst)] ^=
		sh2->R[i_inst_4_7(sh2->inst)];
	sh2->PC += 2;
}

void sh2_opcode_xori(sh2_processor *sh2)
{
	sh2->R[0] ^= i_inst_0_7(sh2->inst);
	sh2->PC += 2;
}

void sh2_opcode_xorm(sh2_processor *sh2)
{
	uint8_t temp;
	
	temp  = sh2_read_byte(sh2, sh2->GBR + sh2->R[0]) ^ i_inst_0_7(sh2->inst);
	
	sh2_write_byte(sh2, sh2->GBR + sh2->R[0], temp);
	sh2->PC += 2;
}

void sh2_opcode_xtrct(sh2_processor *sh2)
{
	uint32_t temp;
	
	temp = (sh2->R[i_inst_4_7(sh2->inst)] << 16) & 0xFFFF0000;
	sh2->R[i_inst_8_11(sh2->inst)] = (sh2->R[i_inst_8_11(sh2->inst)] >> 16) & 0x0000FFFF;
	sh2->R[i_inst_8_11(sh2->inst)] |= temp;
	sh2->PC += 2;
}
