/*
    semu - Sega Saturn Emulator
    Copyright (C) 1998  Chuck Mason

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Chuck Mason IV                      chuckjr@sundail.net
    477 Moffat Loop
    Oviedo, Fl 32765
*/

#include <config.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>

#include "main.h"
#include "header.h"
#include "game.h"
#include "bios.h"
#include "error.h"
#include "gpl.h"

saturn_game current_game;
int show_cdrom_info = 0;
int user_cddevice = 0;
int play_music = 0;
int use_romfile = 0;
int use_cdrom = 0;
int use_sh2_debugger = 0;
int show_version = 0;

char *program_name;

void crtlc(int signal)
{
	printf("Break Caught on signal %d...Exiting...\n", signal);

#ifdef ENABLE_CDROM
	if(play_music)
		ioctl(saturn_cdrom, 0x5307); /* Stop drive motor */
#endif

	if(use_romfile == 0)	
		close(saturn_cdrom);

	exit(1);
}

int main(int argc, char *argv[])
{
	signal(SIGKILL, crtlc);
	signal(SIGTERM, crtlc);
	signal(SIGINT,  crtlc);	

	program_name = (char *)strdup(argv[0]);

	intro();

	if(parse_options(argc, argv) != RET_SUCCESS)
		return RET_FAILURE;

	if(show_version)
		return RET_SUCCESS;
	
	if((use_sh2_debugger == 0) && 
	   ((use_cdrom == 0 && use_romfile == 0) || 
	    (use_cdrom == 1 && use_romfile == 1))) 
		return error("You must specify either -c _or_ -l.\n");

	return emulate_bios((user_cddevice == 1) ? cddevice : DEFAULT_CDDEV);
}

int parse_options(int argc, char *argv[])
{
	int i;
	char *ptr;
	
	for(i = 1; i < argc; i++) {
		if(argv[i][0] == '-') {
			ptr = &argv[i][1];
			while(*ptr) {
				switch(*ptr) {
                                        case 'h':
                                                help();
                                                return RET_FAILURE;
#ifdef ENABLE_CDROM
					case 'c':
						use_cdrom = 1;
						break;
					case 'i':
						show_cdrom_info = 1;
						break;
					case 'd':
						if((i + 1) == argc) {
							help();
							return RET_FAILURE;
						}
						user_cddevice = 1;
						cddevice = (char *)strdup(argv[++i]);
						break;
					case 'm':
						if((i + 1) == argc) {
							help();
							return RET_FAILURE;
						}
						play_music = atoi(argv[++i]);
						break;
#endif					
#ifdef ENABLE_DEBUG
					case '2':
						use_sh2_debugger = 1;
						break;	
#endif
					case 'l':
						if((i + 1) == argc) {
							help();
							return RET_FAILURE;
						}				
						use_romfile = 1;
						romfile = (char *)strdup(argv[++i]);
						break;
					case 'v':
						show_version = 1;
						break;
					default:
						return error(ERR_UNKNOWN_OPTION, *ptr);
				}
				ptr++;
			}
		} else {
			help();
			return RET_FAILURE;
		}
	}
	return RET_SUCCESS;
}
					
void help()
{
#if defined(ENABLE_DEBUG) && defined(ENABLE_CDROM)
	printf("Usage: %s [-<2cmhiv>] <-d device> <-l file>\n", program_name);
	printf("  -d device  -  Set the CD Device to <device>. Default is /dev/cdrom\n");
	printf("  -i         -  Show CDROM header and quit\n");
	printf("  -m track   -  Start playing CDROM music at <track> (CD Audio)\n");
	printf("  -l file    -  Load [file]\n");
	printf("  -c         -  Load CDROM application.\n");
	printf("  -2         -  Enter the interactive SH2 Debugger after ROM load\n");
#endif
#if !defined(ENABLE_DEBUG) && defined(ENABLE_CDROM)
	printf("Usage: %s [-<cmhiv>] <-d device> <-l file>\n", program_name);
	printf("  -d device  -  Set the CD Device to <device>. Default is /dev/cdrom\n");
	printf("  -i         -  Show CDROM header and quit\n");
	printf("  -m track   -  Start playing CDROM music at <track> (CD Audio)\n");
	printf("  -l file    -  Load [file]\n");
	printf("  -c         -  Load CDROM application.\n");
#endif	
#if defined(ENABLE_DEBUG) && !defined(ENABLE_CDROM)
	printf("Usage: %s [-<2hv>] <-l file>\n", program_name);
	printf("  -l file    -  Load [file] into SH2 memory\n");
	printf("  -2         -  Enter the interactive SH2 Debugger after ROM load\n");
#endif
#if !defined(ENABLE_DEBUG) && !defined(ENABLE_CDROM)
	printf("Usage: %s [-<hv>] <-l file>\n", program_name);
	printf("  -l file    -  Load [file] into SH2 memory\n");	
#endif	
	printf("  -h         -  Show this help\n");
	printf("  -v         -  Show version information only\n");
#ifdef ENABLE_CDROM
	printf("\n");
	printf("You must either use -c or -l. semu will not run anything without one.\n");
	printf("\n");
#endif
	return;

}

void intro()
{
	printf("[1;31ms[1;33me[1;34mm[1;37mu[0;37m v" SEMU_VERSION " Copyright (C) 1998 Chuck Mason\n");
	printf("semu compiled on %s at %s\n", __DATE__, __TIME__);
#ifdef ENABLE_DEBUG
	printf("semu comes with ABSOLUTELY NO WARRANTY; for details type `show warranty'\nin the debugger.");
	printf(" This is free software, and you are welcome to redistribute it\n");
	printf("under certain conditions; type `show copying' in the debugger for details.\n");
#else /* !ENABLE_DEBUG */
	printf("semu comes with ABSOLUTELY NO WARRANTY; for details read GPL.\n");
	printf("This is free software, and you are welcome to redistribute it\n");
	printf("under certain conditions; read GPL for details.\n");
#endif /* !ENABLE_DEBUG */
	printf("Use -h for help\n\n");
		                
	return;
}

char *fmttitle(char *s)
{
	int i = strlen(s);
	i -= 1;
	
	for(; i >= 0; i--) {
		if(s[i] != ' ')
			break;
	}
	s[i+1] = '\0';
	
	return s;
}

char *strnchr(char *s, char c, int len)
{
	int i = 0;
	char *ptr;
	
	ptr = s;
	
	while(*ptr != c && i++ < len) 
		ptr++;
	
	if((*ptr == c) && (i < len))
		return ptr;

	return NULL;
}

char *strdate(char *str)
{
        int month, day, year;
        char smonth[3], sday[3], syear[5];
        char date[15];
                        
	char *months[] = {
	 	"Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct",
	 	"Nov", "Dec" 
	};

	syear[0] = str[0];
	syear[1] = str[1];
	syear[2] = str[2];
	syear[3] = str[3];
	syear[4] = '\0';
	smonth[0] = str[4];
	smonth[1] = str[5];
	smonth[2] = '\0';
	sday[0] = str[6];
	sday[1] = str[7];
	sday[2] = '\0';
	
	month = atoi(smonth);
	year  = atoi(syear);
	day   = atoi(sday);
	
	sprintf(date, "%s %d, %d", months[month-1], day, year);
	
	return (char *)strdup(date);
}

